﻿// $Id: nextpleaseOverlay.js,v 1.4 2010/02/07 Alexey Romanov $

// This code was stolen from brody on the mozillazine.org forums.
nextplease.loadListener = {
    onLoad: function (event) {
        getBrowser().addEventListener("DOMContentLoaded", function (event) {
            nextplease.loadListener.onContentLoaded(event);
        },
        true);
    },

    onContentLoaded: function (event) {
        var doc = event.originalTarget;

        // When this becomes true it means that 
        // all of the top level document's 
        // subframes have also finished loading 
        if (!this.isTopLevelDocument(doc)) {return;}

        // Log/alert 
        if (nextplease.DEBUG_DETAILED) {
            var eol = this.getEol();
            var message = "DOMContentLoaded" + eol + doc.title + eol + doc.documentURI;
            nextplease.logDetail("loadListener" + eol + message);
        }
        
        nextplease.cacheImageLocations();
    },

    isTopLevelDocument: function (aDocument) {
        var browsers = getBrowser().browsers;
        for (var i = 0; i < browsers.length; i++) {
            if (aDocument === browsers[i].contentDocument) {return true;}
        }
        return false;
    },

    getPlatform: function () {
        var platform = navigator.platform.toLowerCase();
        if (platform.indexOf('win') !== -1) {return 'win';} // win
        else if (platform.indexOf('mac') !== -1) {return 'mac';} // mac 
        else {return 'other';} // other (nix) 
    },
    
    getEol: function () {
        switch (this.getPlatform()) {
        case 'win':
            return '\r\n';
        case 'mac':
            return '\r';
        case 'other':
            return '\n';
        default:
            return '<eol>';
        }
    }
};

// Read the phrases that specify a next link
// by reading the preferences or defaults,
// and put the phrases in a lookup table.
nextplease.initializeNextPrevMap = function (prefbranch, DefaultArray) {
    // this.logDetail("initializing " + prefbranch);
    var map = {};
    var tempprefname = prefbranch + '.expr0';
    if (this.prefs.prefHasUserValue(tempprefname)) {
        var k = 0;
        while (this.prefs.prefHasUserValue(tempprefname)) {
            //    var phrase = this.prefs.getCharPref(tempprefname);
            var prefValue = this.GetUnicharPref(tempprefname);
            var values = prefValue.split("|");
            for (var j = 0; j < values.length; j++) {
                var phrase = values[j].replace(/&pipe;/g, "|");
                if (phrase !== '') {
                    map[phrase] = true;
                }
            }
            k++;
            tempprefname = prefbranch + '.expr' + k;
        }
    } else {
        for (var i = 0; i < DefaultArray.length; i++) {
            map[DefaultArray[i]] = true;
        }
    }
    // this.logDetail("finished initializing " + prefbranch);
    return map;
};

nextplease.cacheImageLocations = function () {
    this.logDetail("caching image location");
    this.logDetail("there are currently " + this.imageLocationArray.length + " image URLs cached");

    var i;
    var deletedVal;
    
    if (this.DEBUG_DETAILED) {
        this.logDetail("checking cache correctness");
        for (i = 0; i < this.imageLocationArray.length; i++) {
            if (! (this.imageLocationMap[this.imageLocationArray[i]] >= 0)) {
                this.logDetail("URL in array, but not in map: " + this.imageLocationArray[i] + " This should never happen!");
            }
        }
        var url;
        for (url in this.imageLocationMap) {
            if (this.imageLocationMap.hasOwnProperty(url) && (this.imageLocationArray.indexOf(url) == -1)) {
                this.logDetail("URL in map, but not in array: " + url + " This should never happen!");
            }
        }
    }
    
    var theDocument = window.content.document;

    var start = new Date();

    var imgElems = theDocument.getElementsByTagName("img");
    if (imgElems) {
        var numberOfImagesToCheck = Math.min(this.MAX_LINKS_TO_CHECK, imgElems.length);
        this.logDetail("there are " + imgElems.length + " <img> elements, checking " + numberOfImagesToCheck + " of them");
        for (i = 0; i < numberOfImagesToCheck; i++) {
            var imgUrl = imgElems[i].src;
            if (! (this.imageLocationMap[imgUrl] >= 0)) {
                // this.logDetail("adding " + imgUrl + " to image cache");
                if (this.imageLocationArray.length >= this.MAX_IMAGE_CACHE) {
                    deletedVal = this.imageLocationArray.shift();
                    if (this.imageLocationMap[deletedVal] >= 0) {
                        delete this.imageLocationMap[deletedVal];
                        // this.logDetail("deleted " + deletedVal + " from image cache");
                    }
                    else {
                        this.log("tried to delete absent element from image cache. This should never happen!");
                    }
                }

                this.imageLocationMap[imgUrl] = this.imageLocationArray.length;
                this.imageLocationArray[this.imageLocationArray.length] = imgUrl;
            }
            //else {
            //    this.logDetail("" + imgUrl + " is already in image cache");
            //}
        }
    }
    
    var links = theDocument.getElementsByTagName("a");
    if (links) {
        var numberOfLinksToCheck = Math.min(this.MAX_LINKS_TO_CHECK, links.length);
        this.logDetail("there are " + links.length + " <a> elements, checking " + numberOfLinksToCheck + " of them");
        for (i = 0; i < numberOfLinksToCheck; i++) {
            var linkUrl = links[i].href;
            if (! (this.imageLocationMap[linkUrl] >= 0)) {
                if (linkUrl.length > 3) {
                    var suffixIndex = linkUrl.length - 3;
                    var suffix = linkUrl.substr(suffixIndex).toLowerCase();

                    // This is technically not correct, but performance sucks
                    // if we do a proper regex check.
                    if ((suffix === 'jpg') || (suffix === 'gif') || (suffix === 'png') || (suffix === 'bmp')) {
                        // this.logDetail("adding " + linkUrl + " to image cache");
                        if (this.imageLocationArray.length >= this.MAX_IMAGE_CACHE) {
                            deletedVal = this.imageLocationArray.shift();
                            if (this.imageLocationMap[deletedVal] >= 0) {
                                delete this.imageLocationMap[deletedVal];
                                // this.logDetail("deleted " + deletedVal + " from image cache");
                            }
                            else {
                                this.log("tried to delete absent element " + deletedVal + " from image cache. This should never happen!");
                            }
                        }
                        
                        this.imageLocationMap[linkUrl] = this.imageLocationArray.length;
                        this.imageLocationArray[this.imageLocationArray.length] = linkUrl;
                    }
                }
            }
            //else {
            //    this.logDetail("" + linkUrl + " is already in image cache");
            //}
        }
    }
    var end = new Date();
    this.logDetail("caching took " + (end.getTime() - start.getTime()) + " ms");

    if (this.DEBUG_DETAILED) {
        this.logDetail("checking cache correctness again");
        for (i = 0; i < this.imageLocationArray.length; i++) {
            if (! (this.imageLocationMap[this.imageLocationArray[i]] >= 0)) {
                this.logDetail("URL in array, but not in map: " + this.imageLocationArray[i] + " This should never happen!");
            }
        }
        var url;
        for (url in this.imageLocationMap) {
            if (this.imageLocationMap.hasOwnProperty(url) && (this.imageLocationArray.indexOf(url) == -1)) {
                this.logDetail("URL in map, but not in array: " + url + " This should never happen!");
            }
        }
    }

    this.logDetail("there are currently " + this.imageLocationArray.length + " image URLs cached");

    // this.logDetail(this.imageLocationArray.toString());
};

nextplease.initModifierString = function (prefname) {
    var modString;
    
    // TODO: Simplify using defaults
    if (this.prefs.prefHasUserValue(prefname)) {
        modString = this.prefs.getCharPref(prefname);
    }
    else {
        // Special case check for setting prev modifier for upgraders
        // for version 0.5.x and earlier
        if ((prefname === 'nextplease.prevkeymodifier') && (this.prefs.prefHasUserValue('nextplease.keymodifier'))) {
            modString = this.prefs.getCharPref('nextplease.keymodifier');
        } else {
            modString = 'ctrl+shift';
        }
    }

    modString = modString.toLowerCase().replace(/\+/g, " ");
    modString = modString.toLowerCase().replace(/ctrl/g, "accel"); // ???

    return modString;
};

nextplease.initKey = function (prefname, keyId, defaultKeyCode, modString) {
    var keyCode;
    var keyString;
    // TODO: Simplify using defaults
    if (this.prefs.prefHasUserValue(prefname)) {
        keyCode = this.prefs.getIntPref(prefname);
    }
    else {
        keyCode = defaultKeyCode;
    }

    var keyElem = document.getElementById(keyId);
    keyElem.setAttribute("modifiers", modString);

    if (this.KeyCodeToNameMap[keyCode]) {
        keyString = this.KeyCodeToNameMap[keyCode];
        keyElem.removeAttribute("key");
        keyElem.setAttribute("keycode", keyString);
    } else {
        keyString = String.fromCharCode(keyCode).toUpperCase();
        keyElem.setAttribute("key", keyString);
        keyElem.removeAttribute("keycode");
    }
    return keyString;
};

// Read the preferences and setup the 
// useCtrl/useShift/useAlt vars and next/prev
// phrase and image lookup tables
nextplease.readPreferences = function () {
    if (!this.prefs) {
        this.log("nextplease.prefs undefined. This shouldn't happen!");
        this.prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    }

    this.DEBUG = this.prefs.getBoolPref("nextplease.log");
    this.DEBUG_DETAILED = this.prefs.getBoolPref("nextplease.log.detailed");

    this.logDetail("reading preferences");

    this.nextModifierString = this.initModifierString('nextplease.keymodifier');
    this.prevModifierString = this.initModifierString('nextplease.prevkeymodifier');
    this.firstModifierString = this.initModifierString('nextplease.firstkeymodifier');
    this.lastModifierString = this.initModifierString('nextplease.lastkeymodifier');

    this.nextKeyString = this.initKey('nextplease.nextkey', 'nextpleasekey', KeyEvent.DOM_VK_RIGHT, this.nextModifierString);
    this.prevKeyString = this.initKey('nextplease.prevkey', 'nextpleaseprevkey', KeyEvent.DOM_VK_LEFT, this.prevModifierString);
    this.firstKeyString = this.initKey('nextplease.firstkey', 'nextpleasefirstkey', KeyEvent.DOM_VK_UP, this.firstModifierString);
    this.lastKeyString = this.initKey('nextplease.lastkey', 'nextpleaselastkey', KeyEvent.DOM_VK_DOWN, this.lastModifierString);

    // this.logDetail("keys read");

    this.useSubmit = this.prefs.getBoolPref('nextplease.allowsubmit');
    this.useNumberShortcuts = this.prefs.getBoolPref('nextplease.allownumbershortcuts');
    this.useContextMenu = this.prefs.getBoolPref('nextplease.allowcontextmenu');
    this.useSmartNext = this.prefs.getBoolPref('nextplease.allowsmartnext');

    // this.logDetail("bools read");

    // TODO: Simplify using defaults
    var nextRegExString = this.DefaultNextRegEx;
    if (this.prefs.prefHasUserValue('nextplease.nextregex')) {nextRegExString = this.GetUnicharPref('nextplease.nextregex');}
    if (this.Trim(nextRegExString) === '') {nextRegExString = this.DefaultNextRegEx;}
    this.isNextRegExp = new RegExp(nextRegExString, "i");

    var prevRegExString = this.DefaultPrevRegEx;
    if (this.prefs.prefHasUserValue('nextplease.prevregex')) {prevRegExString = this.GetUnicharPref('nextplease.prevregex');}
    if (this.Trim(prevRegExString) === '') {prevRegExString = this.DefaultPrevRegEx;}
    this.isPrevRegExp = new RegExp(prevRegExString, "i");

    var firstRegExString = this.DefaultFirstRegEx;
    if (this.prefs.prefHasUserValue('nextplease.firstregex')) {firstRegExString = this.GetUnicharPref('nextplease.firstregex');}
    if (this.Trim(firstRegExString) === '') {firstRegExString = this.DefaultFirstRegEx;}
    this.isFirstRegExp = new RegExp(firstRegExString, "i");

    var lastRegExString = this.DefaultLastRegEx;
    if (this.prefs.prefHasUserValue('nextplease.lastregex')) {lastRegExString = this.GetUnicharPref('nextplease.lastregex');}
    if (this.Trim(lastRegExString) === '') {lastRegExString = this.DefaultLastRegEx;}

    this.isLastRegExp = new RegExp(lastRegExString, "i");

    // this.logDetail("regexes read");

    var galleryRegExString = this.DefaultGalleryRegEx;
    if (this.prefs.prefHasUserValue('nextplease.galleryregex')) {
        galleryRegExString = this.GetUnicharPref('nextplease.galleryregex');
        if (this.Trim(galleryRegExString) === '') {galleryRegExString = this.DefaultGalleryRegEx;}
        this.galleryUrlRegex = new RegExp(galleryRegExString, "i");
        var matches = this.galleryUrlRegex.exec('http://nextplease.mozdev.org/test/test101.jpg');
        if (matches && (matches.length !== 4)) {
            galleryRegExString = this.DefaultGalleryRegEx;
        }
    }

    this.galleryUrlRegex = new RegExp(galleryRegExString, "i");

    // this.logDetail("gallery regex read");

    this.DisableConflictingKeys();

    var numberkey;
    for (var i = 1; i < 10; i++) {
        numberkey = document.getElementById("nextplease" + i + "key");
        numberkey.setAttribute("disabled", !this.useNumberShortcuts);
    }

    this.NextPhrasesMap = this.initializeNextPrevMap('nextplease.nextphrase', this.DefaultNextPhrases);
    this.PrevPhrasesMap = this.initializeNextPrevMap('nextplease.prevphrase', this.DefaultPrevPhrases);
    this.FirstPhrasesMap = this.initializeNextPrevMap('nextplease.firstphrase', this.DefaultFirstPhrases);
    this.LastPhrasesMap = this.initializeNextPrevMap('nextplease.lastphrase', this.DefaultLastPhrases);

    this.NextImagesMap = this.initializeNextPrevMap('nextplease.nextimage', this.DefaultNextImages);
    this.PrevImagesMap = this.initializeNextPrevMap('nextplease.previmage', this.DefaultPrevImages);
    this.FirstImagesMap = this.initializeNextPrevMap('nextplease.firstimage', this.DefaultFirstImages);
    this.LastImagesMap = this.initializeNextPrevMap('nextplease.lastimage', this.DefaultLastImages);

    this.logDetail("options read");
};

nextplease.DisableKey = function (modifier, keyString) {
    var conflictingKeys;
    if (keyString.indexOf("VK_") >= 0) {
        // this.logDetail("disabling keys conflicting with " + modifier + "+" + this.KeyCodeToNameMap[keycode]);
        conflictingKeys = document.getElementsByAttribute("keycode", keyString);
    }
    else {
        // this.logDetail("disabling keys conflicting with " + modifier + "+" + key);
        conflictingKeys = document.getElementsByAttribute("key", keyString.toLowerCase());
    }

    var conflictingKey, conflictingId, conflictingModifier;
    for (var i = 0; i < conflictingKeys.length; i++) {
        conflictingKey = conflictingKeys[i];
        conflictingId = conflictingKey.getAttribute("id");
        if (! (/nextplease/.test(conflictingId)) && conflictingKey.hasAttribute("modifiers")) {
            conflictingModifier = conflictingKey.getAttribute("modifiers");
            // this.logDetail("potentially conflicting key: " + conflictingId);
            if ((/alt/.test(modifier) === /alt/.test(conflictingModifier)) &&
                (/shift/.test(modifier) === /shift/.test(conflictingModifier)) &&
                (/accel/.test(modifier) === /accel/.test(conflictingModifier))) {
                conflictingKey.parentNode.removeChild(conflictingKey);
                // this.logDetail("actually conflicting");
            }
            else {
                // this.logDetail("not actually conflicting");
            }
        }
    }
};

nextplease.DisableConflictingKeys = function () {
    // disables any conflicting key-bindings
    var conflictingKeys;
    this.logDetail("disabling conflicting keys");

    // Check for ALT+1/2/3... key bindings
    if (this.useNumberShortcuts) {
        var accelKey = this.prefs.getIntPref('ui.key.accelKey');
        for (var i=1; i<10; i++) {
            this.DisableKey("alt", "" + i);
            // Check if ACCEL = ALT?
            // 18 is the ALT key
            if (accelKey === 18) {
                this.DisableKey("accel", "" + i);
            }
        }
    }

    this.DisableKey(this.nextModifierString, this.nextKeyString);
    this.DisableKey(this.prevModifierString, this.prevKeyString);
    this.DisableKey(this.firstModifierString, this.firstKeyString);
    this.DisableKey(this.lastModifierString, this.lastKeyString);
};

nextplease.init = function () {
    this.log("initializing");
    this.readPreferences();
};

// Looks through all the links on the page
// and tries to look for one whose text matches
// one of the phrases or images. If so, it goes to/
// the corresponding link. 
// pages with frames.
nextplease.getLink = function (curWindow, phraseMap, imageMap, phraseRegEx, searchType) {
    var text;
    var isInt = /^\s*\[?\s*(\d+)\s*\]?\s*$/;
    var useFrames = true;

    var tmpPageNumLink = null;
    var firstPageNumLink = null;
    var firstPageNum;
    var lastPageNumLink = null;
    var nextPageNumLink = null;
    var prevPageNumLink = null;

    var currentPageNum = 100000; // Init to arbitrarily large num
    var tmpPageNum = 100000; // Init to arbitrarily large num
    var isFirstNumber = true;
    var greatestNum = 1;

    var i;
    var link;

    this.logDetail("looking for a link");

    this.readPreferences();

    var range = document.createRange();

    if (this.useSmartNext) {
        if ((searchType === this.NEXT_SEARCH_TYPE) && (getBrowser().canGoForward)) {
            this.log("forward in history");
            return curWindow.history.forward();
        } else if ((searchType === this.PREV_SEARCH_TYPE) && (getBrowser().canGoBack)) {
            this.log("back in history");
            return curWindow.history.back();
        }
    }

    // Look for <LINK> tags
    this.logDetail("checking <link> tags");
    var linktags = curWindow.document.getElementsByTagName("link");

    for (i = 0; i < linktags.length; i++) {
        link = linktags[i];

        // Look for rel attributes for next/prev/first/last
        // 
        if (link.rel && link.href) {
            // Hack to workaround PHPBB's incorrect use of LINK tags.
            if (curWindow.location.href.match("viewtopic.php")) {
                // do nothing
            } else if ((searchType === this.NEXT_SEARCH_TYPE) && (link.rel.toLowerCase() === "next")) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.PREV_SEARCH_TYPE) && ((link.rel.toLowerCase() === "prev") || (link.rel.toLowerCase() === "previous"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.FIRST_SEARCH_TYPE) && ((link.rel.toLowerCase() === "start") || (link.rel.toLowerCase() === "first"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.LAST_SEARCH_TYPE) && ((link.rel.toLowerCase() === "end") || (link.rel.toLowerCase() === "last"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            }
        }
    }

    // Look for <A HREF="..."> tags
    this.logDetail("checking <a> tags");
    var alinks = curWindow.document.links;

    // Search through each link
    for (i = 0; i < alinks.length; i++) {
        link = alinks[i];
        // Look for rel attributes for next/prev/first/last
        if (link.rel && link.href) {
            if ((searchType === this.NEXT_SEARCH_TYPE) && (link.rel.toLowerCase() === "next")) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.PREV_SEARCH_TYPE) && ((link.rel.toLowerCase() === "prev") || (link.rel.toLowerCase() === "previous"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.FIRST_SEARCH_TYPE) && ((link.rel.toLowerCase() === "start") || (link.rel.toLowerCase() === "first"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            } else if ((searchType === this.LAST_SEARCH_TYPE) && ((link.rel.toLowerCase() === "end") || (link.rel.toLowerCase() === "last"))) {
                this.log('found rel="' + link.rel.toLowerCase() + '": ' + link.href);
                return this.openLink(curWindow, link);
            }
        }
        range.selectNode(link);
        text = this.Trim(range.toString());

        // If the text is in the lookup table,
        // go to the URL
        if (phraseMap[text] === true) {
            if (link.href.indexOf("/dictionary") < 0) {
                this.log('found phrase match for "' + text + '": ' + link.href);
                return this.openLink(curWindow, link);
            }
        } else if (link.title && phraseMap[link.title] === true) {
            this.log('found phrase match for "' + link.title + '": ' + link.href);
            return this.openLink(curWindow, link);
        } else {
            // If we don't have an exact match, 
            // try a regular expression test.
            // alert(text);
            if (phraseRegEx.test(text)) {
                if (link.href.indexOf("/dictionary") < 0) {
                    this.log('found regex match for "' + text + '": ' + link.href);
                    return this.openLink(curWindow, link);
                }
            } else if (link.title && phraseRegEx.test(link.title)) {
                this.log('found regex match for "' + link.title + '": ' + link.href);
                return this.openLink(curWindow, link);
            }

            // See if there's an image tag
            var imgElems = link.getElementsByTagName("img");
            if (imgElems.length > 0) {
                this.logDetail("checking images inside <a>...</a>");
                // If the image matches, go to the URL.
                //alert(imgElems[0].src);
                var imageElem = imgElems[0];
                var imgtext = imageElem.alt ? imageElem.alt : imageElem.title;

                if (imageMap[imageElem.src] || (phraseMap[imgtext] === true) || phraseRegEx.test(imgtext)) {
                    this.log('found image match: ' + imageElem.src);
                    return this.openLink(curWindow, link);
                }
            }
        }

        var intMatches = isInt.exec(text);
        if (intMatches) {
            var linkPageNum = parseInt(intMatches[1], 10);
            // If the number is greater than a this.MAX_LINK_NUM
            // it probably doesn't have anything to do with
            // a next/prev link.
            //    alert(linkPageNum);
            if (linkPageNum < this.MAX_LINK_NUM) {

                this.logDetail("found a number link, checking...");
                // Try to figure out what the current page and
                // next/prev links are for pages that just have
                // numbered links like  1 2 x 4 5. 
                //     if (linkPageNum === 1) {
                // We're seeing a number link that is smaller
                // than a previous one so assume that we're
                // starting a new set of number links, and
                // count from the beginning.
                if (linkPageNum <= tmpPageNum) {
                    //    alert(linkPageNum);
                    //    alert(currentPageNum);
                    firstPageNumLink = link;
                    firstPageNum = linkPageNum;
                    greatestNum = linkPageNum;

                    prevPageNumLink = null;
                    nextPageNumLink = null;
                    lastPageNumLink = null;
                    currentPageNum = linkPageNum;

                    firstPageNumLink = link;
                    greatestNum = linkPageNum;
                    lastPageNumLink = null;
                    //} else if (currentPageNum === linkPageNum) {
                    //    currentPageNum++;
                    //    prevPageNumLink = link;
                } else if (tmpPageNum + 1 === linkPageNum) {
                    lastPageNumLink = link;
                } else if (tmpPageNum + 2 === linkPageNum) {
                    prevPageNumLink = tmpPageNumLink;
                    nextPageNumLink = link;
                    lastPageNumLink = link;
                }

                tmpPageNum = linkPageNum;
                tmpPageNumLink = link;
            }
        }
    }

    // next and prev are null so that means
    // we have a solid block of number 3,4,5,6,...
    if ((nextPageNumLink === null) && (prevPageNumLink === null)) {

        // If we start with 1, we're probably on last page.
        // Set prev to be lastPage
        if (firstPageNum === 1) {
            prevPageNumLink = lastPageNumLink;
            lastPageNumLink = null;
        }
        // If we start with 2, we're probably on first page.
        // Set next to be first page
        if (firstPageNum === 2) {
            nextPageNumLink = firstPageNumLink;
            firstPageNumLink = null;
        }
        if (firstPageNum > 2) {
            nextPageNumLink = firstPageNumLink;
            prevPageNumLink = lastPageNumLink;
        }
    }

    if (firstPageNumLink && lastPageNumLink && (firstPageNumLink.text === lastPageNumLink.text)) {
        firstPageNumLink = null;
    }

    this.logDetail("first page seems to be " + firstPageNumLink);
    this.logDetail("previous page seems to be " + prevPageNumLink);
    this.logDetail("next page seems to be " + nextPageNumLink);
    this.logDetail("last page seems to be " + lastPageNumLink);
    // Try to find a match using our number algorithm
    if ((searchType === this.NEXT_SEARCH_TYPE) && nextPageNumLink) {return this.openLink(curWindow, nextPageNumLink);}
    else if ((searchType === this.PREV_SEARCH_TYPE) && prevPageNumLink) {return this.openLink(curWindow, prevPageNumLink);}
    else if ((searchType === this.FIRST_SEARCH_TYPE) && firstPageNumLink) {return this.openLink(curWindow, firstPageNumLink);}
    else if ((searchType === this.LAST_SEARCH_TYPE) && lastPageNumLink) {return this.openLink(curWindow, lastPageNumLink);}

    // Otherwise try looking for next/prev submit buttons 
    // if the user allows it.
    if (this.useSubmit && this.getForm(phraseMap, phraseRegEx)) {return true;}

    // See if we can increment the URL to get to next/prev/first
    var galleryLink = this.getGalleryNumberLink(curWindow, searchType);
    // alert(galleryLink);
    // alert(this.imageLocationArray.length);
    if (galleryLink) {curWindow.open(galleryLink, "_self", "");}

    // None of it worked, so make a recursive call to
    // this.getLink on the frame windows.
    var frames = curWindow.frames;
    for (var j = 0; j < frames.length; j++) {
        if (this.getLink(frames[j], phraseMap, imageMap, phraseRegEx, searchType)) {return true;}
    }
    return false;
};

nextplease.getGalleryNumberLink = function (curWindow, searchType) {
    this.logDetail("trying to change the URL by a suitable number");
    var i;
    // alert(this.imageLocationArray);
    var matches = this.galleryUrlRegex.exec(unescape(curWindow.location.href));
    var prefixUrl;
    var suffixUrl;
    var curNumber;
    var urlNumber;
    var padStr;
    var linkUrl;
    
    if (matches && (matches.length === 4)) {
        prefixUrl = matches[1];
        suffixUrl = matches[3];
        this.logDetail("URL prefix is " + prefixUrl + ", URL suffix is " + suffixUrl);
        if (searchType === this.NEXT_SEARCH_TYPE) {
            curNumber = parseInt(matches[2], 10);
            for (i = 1; i < this.MAX_GALLERY_GAP; i++) {
                urlNumber = curNumber + i;
                padStr = this.padNumber(matches[2], urlNumber);
                linkUrl = prefixUrl + padStr + suffixUrl;
                if (this.imageLocationMap[linkUrl] >= 0) {
                    this.log("gallery URL found: " + linkUrl);
                    return linkUrl;
                }
            }
            urlNumber = curNumber + 1;
            padStr = this.padNumber(matches[2], urlNumber);
            linkUrl = prefixUrl + padStr + suffixUrl;
            return linkUrl;
        } else if (searchType === this.PREV_SEARCH_TYPE) {
            curNumber = parseInt(matches[2], 10);
            for (i = 1; (i < this.MAX_GALLERY_GAP) && (i <= curNumber); i++) {
                urlNumber = curNumber - i;
                padStr = this.padNumber(matches[2], urlNumber);
                linkUrl = prefixUrl + padStr + suffixUrl;
                if (this.imageLocationMap[linkUrl] >= 0) {
                    this.log("gallery URL found: " + linkUrl);
                    return linkUrl;
                }
            }
            urlNumber = curNumber - 1;
            if (urlNumber >= 0) {
                padStr = this.padNumber(matches[2], urlNumber);
                linkUrl = prefixUrl + padStr + suffixUrl;
                this.log("gallery URL found: " + linkUrl);
                return linkUrl;
            }
        } else if (searchType === this.FIRST_SEARCH_TYPE) {
            urlNumber = 1;
            padStr = this.padNumber(matches[2], urlNumber);
            linkUrl = prefixUrl + padStr + suffixUrl;
            this.log("gallery URL found: " + linkUrl);
            return linkUrl;
        }
    }
    return undefined;
};

nextplease.padNumber = function (origNumString, newNum) {
    var padStr = "" + newNum;
    var padLen = origNumString.length - padStr.length;
    for (var i = 0; i < padLen; i++) {
        padStr = "0" + padStr;
    }
    return padStr;
};

nextplease.openLink = function (curWindow, linkNode) {
    if (!linkNode) {
        this.log("tried to open undefined link, this should never happen!");
        return false;
    }
    // If it's got an onclick attr, then try to 
    // simulate a mouse click to activate link.
    if (linkNode.hasAttribute("onclick")) {
        // alert(linkNode.getAttribute("onclick"));
        var e = document.createEvent("MouseEvents");
        // e.initMouseEvent("click", 1, 1, window, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, linkNode);

        // From https://developer.mozilla.org/en/DOM/event.initMouseEvent
        e.initMouseEvent("click", true, true, window,
            0, 0, 0, 0, 0, false, false, false, false, 0, null);
        linkNode.dispatchEvent(e);
    } else {
        curWindow.open(linkNode.href, "_self", "");
    }
    return true;
};

// Look through all the HTML inputs for submit buttons
// that have a value that matches our phrases. If it
// finds a match, it submits the corresponding form
// (which will hopefully take us to the next/prev page).
nextplease.getForm = function (phraseMap, phraseRegEx) {
    this.logDetail("looking for submit buttons");
    var text;

    // Probably would be a little faster to
    // only check forms, but I'm getting problems
    // with them. I'm not sure if it's only on
    // malformed HTML pages, or if it's a Firefox bug.
    var inputs = window.content.document.getElementsByTagName("input");

    for (var i = 0; i < inputs.length; i++) {
        if (inputs[i].type === 'submit') {
            text = this.Trim(inputs[i].value);
            if (phraseMap[text] === true) {
                inputs[i].form.submit();
                return true;
            } else {
                // If we don't have an exact match, 
                // try a quick regular expression test.
                if (phraseRegEx.test(text)) {
                    inputs[i].form.submit();
                    return true;
                }
            }
        }
    }
    return false;
};

// Looks through all the links and finds the link
// that matches the linkNum (an integer between
// 1 and 9). If it finds a match, it will go to 
// that link.
nextplease.getNumberedLink = function (curWindow, linkNum) {
    this.logDetail("looking for a link numbered " + linkNum);
    var text;
    var isInt = /^\s*\[?\s*(\d+)\s*\]?\s*$/;
    var useFrames = true;

    this.readPreferences();

    var alinks = curWindow.document.links;

    // Search through each link
    for (var i = alinks.length - 1; i >= 0; i--) {
        var link = alinks[i];

        text = this.Trim(link.text);

        var intMatches = isInt.exec(text);
        if (intMatches) {
            var linkPageNum = parseInt(intMatches[1], 10);
            if (linkPageNum === linkNum) {
                return this.openLink(curWindow, link);
                //curWindow.open(link.href, "_self", "");
                //return true;
            }
        }
    }

    var frames = curWindow.frames;
    for (var j = 0; j < frames.length; j++) {
        if (this.getNumberedLink(frames[j], linkNum)) {return true;}
    }

    return false;
};

nextplease.getNextLink = function () {
    return this.getLink(window.content, this.NextPhrasesMap, this.NextImagesMap, this.isNextRegExp, this.NEXT_SEARCH_TYPE);
};

nextplease.getPrevLink = function () {
    return this.getLink(window.content, this.PrevPhrasesMap, this.PrevImagesMap, this.isPrevRegExp, this.PREV_SEARCH_TYPE);
};

nextplease.getFirstLink = function () {
    return this.getLink(window.content, this.FirstPhrasesMap, this.FirstImagesMap, this.isFirstRegExp, this.FIRST_SEARCH_TYPE);
};

nextplease.getLastLink = function () {
    return this.getLink(window.content, this.LastPhrasesMap, this.LastImagesMap, this.isLastRegExp, this.LAST_SEARCH_TYPE);
};

nextplease.showHideMenuItems = function () {
    if (gContextMenu) {
        gContextMenu.showItem("nextplease.topMenu", this.useContextMenu);
        gContextMenu.showItem("nextPleasePopupMenu", this.useContextMenu);
        gContextMenu.showItem("prevPleasePopupMenu", this.useContextMenu);
        gContextMenu.showItem("firstPleasePopupMenu", this.useContextMenu);
        gContextMenu.showItem("lastPleasePopupMenu", this.useContextMenu);

        gContextMenu.showItem("nextPleaseAddPrevImage", this.useContextMenu && gContextMenu.onImage);
        gContextMenu.showItem("nextPleaseAddNextImage", this.useContextMenu && gContextMenu.onImage);
        gContextMenu.showItem("nextPleaseAddFirstImage", this.useContextMenu && gContextMenu.onImage);
        gContextMenu.showItem("nextPleaseAddLastImage", this.useContextMenu && gContextMenu.onImage);

        var showPhraseMenu = !gContextMenu.onImage && gContextMenu.onLink;
        
        gContextMenu.showItem("nextPleaseAddNextPhrase", this.useContextMenu && showPhraseMenu);
        gContextMenu.showItem("nextPleaseAddPrevPhrase", this.useContextMenu && showPhraseMenu);
        gContextMenu.showItem("nextPleaseAddFirstPhrase", this.useContextMenu && showPhraseMenu);
        gContextMenu.showItem("nextPleaseAddLastPhrase", this.useContextMenu && showPhraseMenu);
    }
};

nextplease.addPhrase = function (map, prefname, defaultValues) {
    this.logDetail("adding a phrase to " + prefname);
    if (gContextMenu.onLink) {
        var range = document.createRange();
        range.selectNode(document.popupNode);
        var text = this.Trim(range.toString());
        this.logDetail("the phrase is: " + text);
        if (text !== '') {
            if ((map !== this.NextPhrasesMap) && (this.NextPhrasesMap[text])) {
                alert('This phrase is already being used for Next.\n' + 'If you want, use the NextPlease! options page to remove it.');
            } else if ((map !== this.PrevPhrasesMap) && (this.PrevPhrasesMap[text])) {
                alert('This phrase is already being used for Prev.\n' + 'If you want, use the NextPlease! options page to remove it.');
            } else if ((map !== this.FirstPhrasesMap) && (this.FirstPhrasesMap[text])) {
                alert('This phrase is already being used for First.\n' + 'If you want, use the NextPlease! options page to remove it.');
            } else if ((map !== this.LastPhrasesMap) && (this.LastPhrasesMap[text])) {
                alert('This phrase is already being used for Last.\n' + 'If you want, use the NextPlease! options page to remove it.');
            } else if (!map[text]) {
                map[text] = true;
                this.addToPrefs(prefname, text, defaultValues);
                this.logDetail("added phrase " + text + " to " + prefname);
            }
        }
    }
    else {
        this.log("tried to add a phrase without popup. This should never happen!");
    }
};

nextplease.addImage = function (map, prefname, defaultValues) {
    this.logDetail("adding an image to " + prefname);
    if (document.popupNode) {
        var imageurl = document.popupNode.src;
        this.logDetail("the image URL is: " + imageurl);
        if ((map !== this.NextImagesMap) && (this.NextImagesMap[imageurl])) {
            alert('This image is already being used for Next.\n' + 'If you want, use the NextPlease! options page to remove it.');
        } else if ((map !== this.PrevImagesMap) && (this.PrevImagesMap[imageurl])) {
            alert('This image is already being used for Prev.\n' + 'If you want, use the NextPlease! options page to remove it.');
        } else if ((map !== this.FirstImagesMap) && (this.FirstImagesMap[imageurl])) {
            alert('This image is already being used for First.\n' + 'If you want, use the NextPlease! options page to remove it.');
        } else if ((map !== this.LastImagesMap) && (this.LastImagesMap[imageurl])) {
            alert('This image is already being used for Last.\n' + 'If you want, use the NextPlease! options page to remove it.');
        } else if (!map[imageurl]) {
            map[imageurl] = true;
            this.addToPrefs(prefname, imageurl, defaultValues);
            this.logDetail("added image " + imageurl + " to " + prefname);
        }
    }
    else {
        this.log("tried to add an image without popup. This should never happen!");
    }
};

nextplease.addToPrefs = function (prefbranch, text, defaultValues) {
    this.logDetail("adding " + text + " to " + prefbranch);
    var tempprefname = prefbranch + '.expr0';
    var prefvalue;
    if (this.prefs.prefHasUserValue(tempprefname)) {
        prefvalue = this.GetUnicharPref(tempprefname);
    } else {
        prefvalue = "";
        for (var i = 0; i < defaultValues.length; i++) {
            prefvalue = prefvalue + "|" + defaultValues[i].replace(/\|/g, "&pipe;");
        }
    }
    text = prefvalue + "|" + text.replace(/\|/g, "&pipe;");

    this.SetUnicharPref(tempprefname, text);
};

window.addEventListener("load", function() { nextplease.init(); }, false);
window.addEventListener("popupshowing", function() { nextplease.showHideMenuItems(); }, false);
window.addEventListener("load",function(event) { nextplease.loadListener.onLoad(event); },false); 
